import os


def pcraster_command(cmd, files=None):

    if files is not None:
        # replace placeholders in command with actual filenames
        for alias, realname in files.items():
            cmd = cmd.replace(alias, '"{}"'.format(realname))

    os.system(cmd)
    return cmd



"""
Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

cutmaps
A tool to cut netcdf files
"""

import argparse
import os
import shutil
import sys

from .. import version
from .cutlib import mask_from_ldd, get_filelist, get_cuts, cutmap, logger


class ParserHelpOnError(argparse.ArgumentParser):
    def error(self, message):
        sys.stderr.write('Error: %s\n' % message)
        self.print_help()
        sys.exit(1)

    def add_args(self):
        group_mask = self.add_mutually_exclusive_group()
        group_filelist = self.add_mutually_exclusive_group()
        group_stations = group_mask.add_argument_group()

        group_mask.add_argument("-m", "--mask", help='mask file cookie-cutter, .map if pcraster, .nc if netcdf')
        group_mask.add_argument("-c", "--cuts", help='Cut coordinates in the form lonmin_lonmax:latmin_latmax')

        group_stations.add_argument("-l", "--ldd", help='Path to LDD file')
        group_stations.add_argument("-N", "--stations", help='Path to stations.txt file.'
                                                             'Read documentation to know about the format')
        group_stations.add_argument("-C", "--clonemap",
                                    help='Path to PCRaster clonemap; used to convert ldd.nc to ldd.map')
        group_mask.add_argument_group(group_stations)

        group_filelist.add_argument("-f", "--folder", help='Directory with netCDF files to be cut')
        group_filelist.add_argument("-S", "--static-data", help='Directory with EFAS/GloFAS static maps. '
                                                                'Output files will have same directories structure')

        self.add_argument("-o", "--outpath", help='path where to save cut files', default='./cutmaps_out',
                          required=True)
        self.add_argument("-W", "--overwrite", help='Set flag to overwrite existing files', default=False,
                          required=False, action='store_true')


def main(cliargs):
    parser = ParserHelpOnError(description='Cut netCDF file: {}'.format(version))
    parser.add_args()
    args = parser.parse_args(cliargs)

    mask = args.mask
    cuts = args.cuts

    ldd = args.ldd
    stations = args.stations

    input_folder = args.folder
    static_data_folder = args.static_data
    overwrite = args.overwrite
    pathout = args.outpath
    if not os.path.exists(pathout):
        logger.warning('\nOutput folder %s not existing. Creating it...', pathout)
        os.mkdir(pathout)
    if ldd and stations:
        logger.info('\nTry to produce a mask from LDD and stations points: %s %s', ldd, stations)
        if ldd.endswith('.nc'):
            from lisfloodutilities.nc2pcr import convert
            clonemap = args.clonemap
            ldd = convert(ldd, clonemap, '.map')
        mask = mask_from_ldd(ldd, stations)

    logger.info('\n\nCutting using: %s\n Files to cut from: %s\n Output: %s\n Overwrite existing: %s\n\n',
                mask or cuts,
                input_folder or static_data_folder,
                pathout, overwrite)

    list_to_cut = get_filelist(input_folder, static_data_folder)
    x_min, x_max, y_min, y_max = get_cuts(cuts=cuts, mask=mask)

    # walk through list_to_cut
    for file_to_cut in list_to_cut:

        filename, ext = os.path.splitext(file_to_cut)

        # localdir used only with lisflood_static_data_folder.
        # It will track folder structures in a EFAS/GloFAS like setup and replicate it in output folder
        localdir = os.path.dirname(file_to_cut)\
            .replace(os.path.dirname(static_data_folder), '')\
            .lstrip('/') if static_data_folder else ''

        fileout = os.path.join(pathout, localdir, os.path.basename(file_to_cut))
        if os.path.isdir(file_to_cut) and static_data_folder:
            # just create folder
            os.makedirs(fileout, exist_ok=True)
            continue
        if ext != '.nc':
            if static_data_folder:
                logger.warning('%s is not in netcdf format, just copying to ouput folder', file_to_cut)
                shutil.copy(file_to_cut, fileout)
            else:
                logger.warning('%s is not in netcdf format, skipping...', file_to_cut)
            continue
        elif os.path.isfile(fileout) and os.path.exists(fileout) and not overwrite:
            logger.warning('%s already existing. This file will not be overwritten', fileout)
            continue

        cutmap(file_to_cut, fileout, x_min, x_max, y_min, y_max)


def main_script():
    sys.exit(main(sys.argv[1:]))


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



"""

Copyright 2019-2020 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""

import os
import logging
import itertools
import datetime

from nine import IS_PYTHON2
if IS_PYTHON2:
    from pathlib2 import Path
else:
    from pathlib import Path

import numpy as np
from netCDF4 import Dataset, date2index

from lisfloodutilities.readers import PCRasterMap


logging.basicConfig(level=logging.INFO, format='[%(levelname)s] %(message)s')
logger = logging.getLogger()


class Comparator(object):
    """

    """
    glob_expr = None

    def __init__(self, array_equal=False, for_testing=False):
        """

        """
        self.array_equal = array_equal
        self.for_testing = for_testing
        self.errors = []

    def compare_dirs(self, path_a, path_b, skip_missing=True, timestep=None):
        logger.info('Comparing %s and %s [skip missing: %s]', path_a, path_b, str(skip_missing))
        path_a = Path(path_a)
        path_b = Path(path_b)
        for fa in itertools.chain(*(path_a.glob(e) for e in self.glob_expr)):
            fb = path_b.joinpath(fa.name)
            if not fb.exists():
                if skip_missing:
                    logger.info('skipping %s as it is not in %s', fb.name, path_b.as_posix())
                    continue
                else:
                    message = '{} is missing in {}'.format(fb.name, path_b.as_posix())
                    if self.for_testing:
                        assert False, message
                    else:
                        self.errors.append(message)
                        continue

            self.compare_files(fa.as_posix(), fb.as_posix(), timestep)
        if not self.for_testing:
            return self.errors

    def compare_files(self, fa, fb, timestep=None):
        raise NotImplementedError()


class PCRComparator(Comparator):
    # TODO add comparison with tolerance
    glob_expr = ['**/*.[0-9][0-9][0-9]', '**/*.map']

    def compare_files(self, file_a, file_b, timestep=None):
        logger.info('Comparing %s and %s', file_a, file_b)
        map_a = PCRasterMap(file_a)
        map_b = PCRasterMap(file_b)

        if map_a != map_b:
            map_a.close()
            map_b.close()
            message = '{} different from {}'.format(file_a, file_b)
            if self.for_testing:
                assert False, message
            else:
                self.errors.append(message)
                return message
        else:
            map_a.close()
            map_b.close()
            assert True


class TSSComparator(Comparator):
    glob_expr = ['**/*.tss']

    def __init__(self, atol=0.0001, rtol=0.001,
                 array_equal=False, for_testing=True):

        super(TSSComparator, self).__init__(array_equal=array_equal, for_testing=for_testing)
        self.atol = atol
        self.rtol = rtol

    def compare_lines_equal(self, file_a, file_b):
        with open(file_a, 'rb') as fp1, open(file_b, 'rb') as fp2:
            # skip first line in TSS as it just reports settings filename
            fp1.readline()
            fp2.readline()

            while True:
                b1 = fp1.readline()
                b2 = fp2.readline()
                if (b1 != b2) or (not b1 and b2) or (not b2 and b1):
                    message = '{} different from {}'.format(file_a, file_b)
                    if self.for_testing:
                        assert False, message
                    else:
                        self.errors.append(message)
                        return message
                if not b1:
                    break

    def compare_lines_tolerance(self, file_a, file_b):
        fp1 = open(file_a)
        fp2 = open(file_b)
        # skip first lines in TSS as it just reports settings filename
        line = ''
        numline = 1
        while 'timestep' not in line:
            fp1.readline()
            line = fp2.readline()
            numline += 1

        while True:
            b1 = fp1.readline().strip().split()
            b2 = fp2.readline().strip().split()
            if not b1:
                break
            array1 = np.array(b1, dtype='float64')
            array2 = np.array(b2, dtype='float64')
            if not np.allclose(array1, array2, rtol=self.rtol, atol=self.atol, equal_nan=True):
                message = '{} different from {} (line: {})\n line A: {}\n line B: {}'.format(file_a, file_b, numline, array1, array2)
                message += '\n {}'.format(np.abs(array1 - array2))
                if self.for_testing:
                    np.testing.assert_allclose(array1, array2, rtol=self.rtol, atol=self.atol, equal_nan=True)
                    assert False, message
                else:
                    self.errors.append(message)
        assert True
        return self.errors

    def compare_files(self, file_a, file_b, timestep=None):
        logger.info('Comparing %s and %s', file_a, file_b)
        if self.array_equal:
            self.compare_lines_equal(file_a, file_b)
        else:
            self.compare_lines_tolerance(file_a, file_b)


class NetCDFComparator(Comparator):
    glob_expr = ['**/*.nc']

    def __init__(self, mask, atol=0.0001, rtol=0.001,
                 max_perc_diff=0.2, max_perc_large_diff=0.1,
                 array_equal=False, for_testing=True):

        super(NetCDFComparator, self).__init__(array_equal=array_equal, for_testing=for_testing)

        if isinstance(mask, str):
            mask = Dataset(mask)
            maskvar = [k for k in mask.variables if len(mask.variables[k].dimensions) == 2][0]
            self.maskarea = np.logical_not(mask.variables[maskvar][:, :])
        else:
            self.maskarea = mask

        self.atol = atol
        self.rtol = rtol
        self.max_perc_large_diff = max_perc_large_diff
        self.max_perc_diff = max_perc_diff
        self.large_diff_th = self.atol * 10

    def compare_arrays(self, vara_step, varb_step, varname=None, step=None, filepath=None):

        vara_step = np.ma.compressed(np.ma.masked_array(vara_step, self.maskarea)).astype('float64')
        varb_step = np.ma.compressed(np.ma.masked_array(varb_step, self.maskarea)).astype('float64')
        diff_values = np.ma.abs(vara_step - varb_step)
        diff_values = diff_values[~np.isnan(diff_values)]
        same_values = np.ma.allclose(diff_values, np.zeros(diff_values.shape), atol=self.atol, rtol=self.rtol)
        all_ok = vara_step.size == varb_step.size and same_values
        array_ok = np.isclose(diff_values, np.zeros(diff_values.shape), atol=self.atol, rtol=self.rtol, equal_nan=True)
        different_values_size = array_ok[~array_ok].size
        if (not all_ok) and (different_values_size > 0):
            max_diff = np.ma.amax(diff_values)  # returns a scalar
            perc_wrong = different_values_size * 100 / vara_step.size
            result = np.ma.where(diff_values >= max_diff)
            rel_diff = max_diff * 100. / np.maximum(vara_step[result], varb_step[result])
            if rel_diff.size > 0 and np.max(rel_diff) > 0.01 and (perc_wrong >= self.max_perc_diff or (perc_wrong >= self.max_perc_large_diff and max_diff > self.large_diff_th)):
                step = step if step is not None else '(no time)'
                filepath = os.path.basename(filepath) if filepath else '<mem>'
                varname = varname or '<unknown var>'
                message = '{}/{}@{} - {:3.2f}% of different values - max diff: {:3.6f}'.format(filepath, varname, step, perc_wrong, max_diff)
                logger.error(message)
                if self.for_testing:
                    assert False, message
                else:
                    self.errors.append(message)
                    return message
        assert True

    def compare_files(self, file_a, file_b, timestep=None):
        logger.info('Comparing %s and %s %s', file_a, file_b, timestep or '')
        if timestep and not isinstance(timestep, datetime.datetime):
            raise ValueError('timestep must be of type datetime.datetime but type {} was found'.format(str(type(timestep))))
        with Dataset(file_a) as nca, Dataset(file_b) as ncb:
            num_dims = 3 if 'time' in nca.variables else 2
            var_name = [k for k in nca.variables if len(nca.variables[k].dimensions) == num_dims][0]
            vara = nca.variables[var_name]
            try:
                varb = ncb.variables[var_name]
            except KeyError:
                var_nameb = [k for k in ncb.variables if len(ncb.variables[k].dimensions) == num_dims][0]
                message = 'Files: {} vs {} have different variables names A:{} B:{}'.format(file_a, file_b, var_name, var_nameb)
                if self.for_testing:
                    assert False, message
                else:
                    self.errors.append(message)
                    return message
            if 'time' in nca.variables:
                if not timestep:
                    stepsa = nca.variables['time'][:]
                    stepsb = ncb.variables['time'][:]
                    if len(stepsa) != len(stepsb):
                        message = 'Files: {} vs {}: different number of steps A:{} B:{}'.format(file_a, file_b, len(stepsa), len(stepsb))
                        if self.for_testing:
                            assert False, message
                        else:
                            self.errors.append(message)
                            return message
                    for step, _ in enumerate(stepsa):
                        values_a = vara[step][:, :]
                        values_b = varb[step][:, :]
                        if not self.array_equal:
                            mess = self.compare_arrays(values_a, values_b, var_name, step, file_a)
                        else:
                            mess = self.compare_arrays_equal(values_a, values_b, var_name, step, file_a)
                        if mess:
                            self.errors.append(mess)
                else:
                    # check arrays at a given timestep
                    ia = date2index(timestep, nca.variables['time'], nca.variables['time'].calendar)
                    ib = date2index(timestep, ncb.variables['time'], ncb.variables['time'].calendar)
                    values_a = vara[ia][:, :]
                    values_b = varb[ib][:, :]
                    if not self.array_equal:
                        mess = self.compare_arrays(values_a, values_b, var_name, ia, file_a)
                    else:
                        mess = self.compare_arrays_equal(values_a, values_b, var_name, ia, file_a)
                    if mess:
                        self.errors.append(mess)
            else:
                values_a = vara[:, :]
                values_b = varb[:, :]
                if not self.array_equal:
                    mess = self.compare_arrays(values_a, values_b, var_name, filepath=file_a)
                else:
                    mess = self.compare_arrays_equal(values_a, values_b, var_name, filepath=file_a)
                if mess:
                    self.errors.append(mess)
            return self.errors

    def compare_arrays_equal(self, values_a, values_b, varname=None, step=None, filepath=None):
        filepath = os.path.basename(filepath) if filepath else '<mem>'
        message = '{}/{}@{} is different'.format(filepath, varname, step)
        if self.for_testing:
            np.testing.assert_array_equal(values_a, values_b, message)
        else:
            if not np.array_equal(values_a, values_b):
                self.errors.append(message)
                return message



"""
Copyright 2019-2020 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

nc2prc
A tool that convert netCDF files to PCRaster maps
"""

import sys
import argparse

from . import convert
from .. import version


class ParserHelpOnError(argparse.ArgumentParser):
    def error(self, message):
        sys.stderr.write('Error: %s\n' % message)
        self.print_help()
        sys.exit(1)

    def add_args(self):
        self.add_argument('-i', '--input', required=True,
                          help='Path to input NetCDF. Current version does not work with mapstacks',
                          metavar='input')
        self.add_argument('-o', '--output', help='Path to PCRaster output map.', default='./out.map',
                          metavar='output')
        self.add_argument('-c', '--clonemap', help='Path to PCRaster clone map.',
                          metavar='clonemap')
        self.add_argument("-l", "--is_ldd", help='Set flag if input file is a LDD map', default=False,
                          required=False, action='store_true')


def main_script():
    sys.exit(main(sys.argv[1:]))


def main(args):
    parser = ParserHelpOnError(description='Convert a netCDF file to a PCRaster map {}'.format(version))

    parser.add_args()
    parsed_args = parser.parse_args(args)
    configuration = {'inf': parsed_args.input,
                     'clonemap': parsed_args.clonemap,
                     'outf': parsed_args.output,
                     'is_ldd': parsed_args.is_ldd}
    # MAIN METHOD
    convert(**configuration)


if __name__ == '__main__':
    main_script()



import xarray as xr

x_coordinates_names = ('lon', 'x', 'longitude')
y_coordinates_names = ('lat', 'y', 'latitude')
coordinates_names = x_coordinates_names + y_coordinates_names


class NetCDFMap:
    """
    A class representing a netCDF map.
    Current version does not support mapstacks.
    e.g. To convert ldd.nc
    """
    def __init__(self, netcdf_filename):
        self.ds = xr.open_dataset(netcdf_filename, decode_cf=False)

    @property
    def mv(self):
        for variable in self.ds.variables.values():
            if len(variable.dims) < 2:
                continue
            return variable.attrs['_FillValue'] if '_FillValue' in variable.attrs else variable.attrs['missing_value']

    @property
    def data(self):
        for varname, variable in self.ds.variables.items():
            if len(variable.dims) < 2:
                continue
            yield varname, variable

    @property
    def coordinates(self):
        res = {}
        for varname, variable in self.ds.variables.items():
            if varname.lower() not in coordinates_names:
                continue
            key = 'x' if varname.lower() in x_coordinates_names else 'y'
            res[key] = variable.values
        return res

    def close(self):
        self.ds.close()



"""
Copyright 2019-2020 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.


---------------------------------------------------------------------------------------------------------------------------------------
To publish a new version of this distribution (git tags and pypi package), after pushed on main branch:

python setup.py testpypi
python setup.py publish

To upload new package on PyPi Test:
twine upload --repository-url https://test.pypi.org/legacy/ dist/*

Test package install
pip install --index-url https://test.pypi.org/simple/ lisflood-utilities==0.11.7

Installation with pip:
pip install lisflood-utilities
"""

import os
import sys
import subprocess
from shutil import rmtree

from setuptools import setup, find_packages, Command

current_dir = os.path.dirname(os.path.abspath(__file__))
readme_file = os.path.join(current_dir, 'README.md')
version_file = os.path.join(current_dir, 'src/lisfloodutilities/VERSION')

with open(readme_file, 'r') as f:
    long_description = f.read()

with open(version_file, 'r') as f:
    version = f.read().strip()


def _get_gdal_version():
    try:
        p = subprocess.Popen(['gdal-config', '--version'], stdout=subprocess.PIPE)
    except FileNotFoundError:
        raise SystemError('gdal-config not found.'
                          'GDAL seems not installed. '
                          'Please, install GDAL binaries and libraries for your system '
                          'and then install the relative pip package. '
                          'If you are using a conda environment: conda install -c conda-forge gdal')
    else:
        return p.communicate()[0].splitlines()[0].decode()


IS_PYTHON2 = sys.version_info.major == 2
numpy_version = '1.18.3' if not IS_PYTHON2 else '1.15.4'
dask_version = '2.7.0' if not IS_PYTHON2 else '1.2.2'
gdal_version = _get_gdal_version()


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Publish lisflood-utilities package.'
    user_options = []

    @staticmethod
    def print_console(s):
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.print_console('Removing previous builds...')
            rmtree(os.path.join(current_dir, 'dist'))
        except OSError:
            pass

        self.print_console('Building Source and Wheel (universal) distribution...')
        os.system('{0} setup.py sdist'.format(sys.executable))

        self.print_console('Uploading the package to PyPI via Twine...')
        os.system('twine upload dist/*')

        self.print_console('Pushing git tags...')
        os.system('git tag v{0}'.format(version))
        os.system('git push --tags')

        sys.exit()


class UploadCommandTest(UploadCommand):

    def run(self):
        try:
            self.print_console('Removing previous builds...')
            rmtree(os.path.join(current_dir, 'dist'))
        except OSError:
            pass

        self.print_console('Building Source and Wheel (universal) distribution...')
        os.system('{} setup.py sdist'.format(sys.executable))

        self.print_console('Uploading the package to test PyPI via Twine...')
        os.system('twine upload --repository testpypi dist/*')

        sys.exit()


setup_args = dict(
    name='lisflood-utilities',
    package_dir={'': 'src/'},
    package_data={'lisfloodutilities': ['VERSION']},
    version=version,
    packages=find_packages('src'),
    description='A set of utilities for lisflood users. '
                'pcr2nc: Convert PCRaster files to netCDF CF 1.6; '
                'nc2pcr: Convert netCDF files ot PCRaster format; '
                'cutmaps: cut netCDF files;'
                'compare: compare two set of netcdf files',
    long_description=long_description,
    long_description_content_type='text/markdown',
    setup_requires=[
            'setuptools>=41.0', 'numpy=={}'.format(numpy_version),
    ],
    install_requires=['numpy=={}'.format(numpy_version), 'pyyaml==5.3',
                      # Can create corrupted environment if using conda,
                      # Better to install GDAL manually before to install lisflood-utilities package
                      # 'GDAL=={}'.format(gdal_version),
                      'netCDF4==1.5.3', 'toolz', 'xarray==0.15.1',
                      'dask=={}'.format(dask_version), 'pandas==0.25.1', 'pathlib2==2.3.5', 'nine'],
    author="Valerio Lorini, Domenico Nappo, Lorenzo Alfieri",
    author_email="valerio.lorini@ec.europa.eu,domenico.nappo@gmail.com,lorenzo.alfieri@ec.europa.eu",
    keywords=['netCDF4', 'PCRaster', 'mapstack', 'lisflood', 'efas', 'glofas', 'ecmwf', 'copernicus'],
    license='EUPL 1.2',
    url='https://github.com/ec-jrc/lisflood-utilities',
    scripts=['bin/pcr2nc', 'bin/cutmaps', 'bin/compare', 'bin/nc2pcr', ],
    zip_safe=True,
    classifiers=[
          # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Intended Audience :: Education',
          'Intended Audience :: Other Audience',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: European Union Public Licence 1.2 (EUPL 1.2)',
          'Operating System :: Unix',
          'Operating System :: POSIX',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: MacOS :: MacOS X',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Topic :: Scientific/Engineering :: Physics',
    ],
    # setup.py publish to pypi.
    cmdclass={
        'publish': UploadCommand,
        'upload': UploadCommand,
        'testpypi': UploadCommandTest,
    },
)

setup(**setup_args)



import os
import uuid

from osgeo import gdal

import numpy as np
import numpy.ma as ma


class PCRasterWriter:
    FORMAT = 'PCRaster'
    dtype_to_valuescale = {
        'uint8': (gdal.GDT_Byte, 'VS_BOOLEAN', np.int8),
        'int8': (gdal.GDT_Int32, 'VS_NOMINAL', np.int32),
        'int16': (gdal.GDT_Int32, 'VS_ORDINAL', np.int32),
        'float32': (gdal.GDT_Float32, 'VS_SCALAR', np.float32),
        'float64': (gdal.GDT_Float32, 'VS_SCALAR', np.float32),
    }

    def __init__(self, *args, **kwargs):
        """
        Constructor accept only keywords arguments:
        Must have either a clonemap or coordinates in input
        clonemap: path to a clone PCRaster map
        coordinates: dict of numpy 1D arrays representing coordinates, with 'x' and 'y' keys
        mv: Missing value (optional)
        """
        self._drv = gdal.GetDriverByName(self.FORMAT)
        self._drv.Register()
        self._clone_map = kwargs.get('clonemap')
        self._coordinates = kwargs.get('coordinates')
        mv = kwargs.get('mv')
        self.flipped_x = False
        self.flipped_y = False
        if not (self._clone_map or self._coordinates):
            raise ValueError('Need either a PCRaster clonemap or coordinates={"x": [...], "y": [...]}')

        if self._clone_map:
            src_ds = gdal.Open(self._clone_map.encode('utf-8'), gdal.GA_ReadOnly)
            src_band = src_ds.GetRasterBand(1)

            # Producing mask array
            self.cols = src_ds.RasterXSize
            self.rows = src_ds.RasterYSize
            self.src_geo_transform = src_ds.GetGeoTransform()

            rs = src_band.ReadAsArray()
            no_data_value = src_band.GetNoDataValue()
            src_ds = None
            del src_ds
            src_band = None
            del src_band
            rs = ma.masked_values(rs, no_data_value, copy=False)
            self._mask = ma.getmask(rs)
            rs = None
            self.mv = no_data_value if not mv else float(mv)
        else:
            self.cols = self._coordinates['x'].shape[0]
            self.rows = self._coordinates['y'].shape[0]
            self.src_geo_transform = self._get_geo_transform_from_coords()
            self.mv = np.nan if not mv else float(mv)
            self.flipped_x = self._coordinates['x'][0] > self._coordinates['x'][1]
            self.flipped_y = self._coordinates['y'][0] < self._coordinates['y'][1]

    def write(self, output_map_name, values, is_ldd=False):
        if is_ldd:
            etype = gdal.GDT_Byte
            valuescale = 'VS_LDD'
            as_type = np.int8
        else:
            etype, valuescale, as_type = self.get_valuescale(values)

        if not self._clone_map:
            # if no clone map was provided, we can calculate mask once we get values
            masked_values = ma.masked_values(values, self.mv, copy=False)
            self._mask = ma.getmask(masked_values)

        filled_values = self._mask_values(values, self.mv)
        filled_values = filled_values.astype(as_type)
        tmpfilename = 'tmp{}.map'.format(uuid.uuid4())
        dest_ds = self._drv.Create(tmpfilename, xsize=self.cols, ysize=self.rows, bands=1,
                                   eType=etype,
                                   options=["PCRASTER_VALUESCALE={}".format(valuescale)])

        # =============================================================================
        # Create a MEM clone of the dest file as Create returns readonly dataset
        # =============================================================================
        mem_drive = gdal.GetDriverByName('MEM')
        mem_drive.Register()
        mem_ds = mem_drive.CreateCopy('mem', dest_ds)
        dest_ds = None
        del dest_ds
        mem_ds.SetGeoTransform(self.src_geo_transform)
        rs = mem_ds.GetRasterBand(1)
        if self.flipped_y:
            # when we do not use a clone map, we need to flip up down values
            # not clear..it should be done always since PCRaster y coordinates are upside down.
            filled_values = np.flipud(filled_values)
        if self.flipped_x:
            # when we do not use a clone map, we need to flip up down values
            # not clear..it should be done always since PCRaster y coordinates are upside down.
            filled_values = np.fliplr(filled_values)
        rs.WriteArray(filled_values)
        # FIXME this set missing value statement is ignored once file is written on disk
        rs.SetNoDataValue(self.mv)
        rs.FlushCache()
        # =============== Write on disk
        out_ds = self._drv.CreateCopy(output_map_name.encode('utf-8'), mem_ds)
        out_ds = None
        del out_ds
        mem_ds = None
        del mem_ds
        os.unlink(tmpfilename)

    def _mask_values(self, values, mv=None):
        if isinstance(values, ma.core.MaskedArray):
            masked = ma.masked_where((self._mask | values.mask), values.data, copy=False)
        else:
            masked = ma.masked_where(self._mask, values, copy=False)
        mv = self.mv if not mv else mv
        masked = ma.filled(masked, mv)
        return masked

    def close(self):
        pass

    def get_valuescale(self, values):
        dtype = str(values.dtype)
        return self.dtype_to_valuescale[dtype]

    def _get_geo_transform_from_coords(self):
        # geotransform is just a tuple e.g.  (2500000.0, 5000.0, 0.0, 5500000.0, 0.0, -5000.0)
        # (-25.25, 0.016666666666666666, 0.0, 72.25, 0.0, -0.016666666666666666) geotransform with 1 arcmin europe maps
        # 0 - xul
        # 1 - w-e pixel resolution
        # 2 - 0.0 (?)
        # 3 - yul
        # 4 - 0.0
        # 5 - n-s pixel resolution (negative value)
        #
        top_left_x = np.round(self._coordinates['x'][0], 2)
        w_e_resolution = self._coordinates['x'][1] - self._coordinates['x'][0]

        top_left_y = np.round(self._coordinates['y'][0], 2)
        n_s_resolution = - (self._coordinates['y'][1] - self._coordinates['y'][0])
        return top_left_x, w_e_resolution, 0.0, top_left_y, 0.0, n_s_resolution
